import pandas as pd
import datetime as dt
import numpy as np

from ContractRigObjects import Rig


# ------ Read Rig Input File ------- #


def listofRigs(rig_file):
    """
        Return list of Rig Objects from .csv file.
    """

    df = pd.read_csv(rig_file, index_col='Name')

    # TODO: Input lines below to ReadRigs.py
    df.rename(columns={'Type': 'type',
                       'Time available': 't_available',
                       'Cost of purchase(mill USD)': 'cost_purchase',
                       'OPEX(month)': 'opex',
                       'Water Depth': 'depth',
                       'HE/BE': 'he_be',
                       'Age': 'age',
                       'Speed(knots)': 'speed',
                       'Sources': 'sources', }, inplace=True)

    # Split date
    date_split = df['t_available'].str.split('.', expand=True)
    df['year_a'] = date_split[1]  # add new columns
    df['month_a'] = date_split[0]
    df['month_a'].replace({'jan': 1, 'feb': 2, 'mar': 3,
                           'apr': 4, 'mai': 5, 'jun': 6,
                           'jul': 7, 'aug': 8, 'sep': 9,
                           'okt': 10, 'nov': 11, 'des': 12}, inplace=True)

    # Create timestamp for t_available
    def convert_to_timestamp(year, month):
        my_date = list()
        for i in range(len(year)):
            if month[i] == 'now':  # redundant, now replaced by jan.20
                my_date.append(dt.date(2020, 1, 1))
            else:
                year[i] = int(year[i])
                month[i] = int(month[i])
                my_date.append(dt.date(2000 + year[i], month[i], 1))
        return my_date

    df['t_available'] = convert_to_timestamp(df['year_a'], df['month_a'])

    # Create Object
    rigs = np.empty(0)
    r_id = 1
    for i in df.index:
        x = df.loc[i]
        rigs = np.append(rigs, Rig(r_id, x.name, x.type, x.t_available,
                                   x.cost_purchase, x.opex, x.he_be,
                                   x.depth, x.age, x.speed))
        r_id = r_id + 1

    return rigs
