# coding=utf-8
# Python encoding problem, add line below. https://www.python.org/dev/peps/pep-0263/

# Python libraries
import numpy as np
import pandas as pd
import datetime as dt
import matplotlib.pyplot as plt
# My imports
import LongestPath as lp
import ContractRigObjects as cs
from ContractGenerator import generateContracts
from ReadRigs import listofRigs
from post_process import getPostProcessing
# from plot_something import plotContractTimeline

# --- START INPUT --- #

# --- Simulation choices --- #
T = 10  # simulation period (years)
nIterations = 1  # number of iterations
my_mode = 0  # mode = 0: rig selection, mode = 1: fleet allocation
my_fleets = 1  # number of fleets we are evaluating

# --- Input for contract scenarios --- #


# --- Variable input dependent on oil price scenario --- #
# Oil price - N(mu,sigma)
p_oil_price = np.array([30., 10.])
# p_oil_price = np.array([60., 15.])
# p_oil_price = np.array([100., 25.])

# Number of contracts(oil price and fleet size) - N(mu,sigma)
p_num_contracts = np.array([[30., 1],  # new value 1
                            [60., 2],  # new value 2
                            [100., 3]  # new value 3
                            ])

# Market - B, GOM, NS, WA
p_market = np.array([0.20, 0.28, 0.22, 0.30])

# Water depth by market
if p_oil_price[0] == 30:
    p_depths = np.array([['depth(m)', 50, 1500, 2000],  # Poil=30
                         ['B', 0.05, 0.6, 0.35],
                         ['GOM', 0.7, 0.25, 0.05],
                         ['NS', 0.6, 0.4, 0.],
                         ['WA', 0.8, 0.15, 0.05]
                         ], dtype=object)
elif p_oil_price[0] == 60:
    p_depths = np.array([['depth(m)', 50, 1500, 2000],  # Poil=60
                         ['B', 0.05, 0.6, 0.35],
                         ['GOM', 0.35, 0.35, 0.30],
                         ['NS', 0.4, 0.6, 0.],
                         ['WA', 0.5, 0.3, 0.2]
                         ], dtype=object)
elif p_oil_price[0] == 100:
    p_depths = np.array([['depth(m)', 50, 1500, 2000],  # Poil=100
                         ['B', 0.05, 0.40, 0.55],
                         ['GOM', 0.3, 0.3, 0.4],
                         ['NS', 0.3, 0.7, 0.],
                         ['WA', 0.4, 0.25, 0.35]
                         ], dtype=object)

# --- Constant independent of scenario --- #

# Scope of contract - development, exploration, appraisal
p_scope = np.array([0.74, 0.17, 0.09])

# Duration of contract - N(mu,sigma)
# mu_duration (days) = b0 + b1*oil_price
p_duration = np.array([112.5, 1.9])

# Revenue of contract
# mu_std = b0 + b1 ln(oil_price)
# betas differ for JU and floater
p_revenue = np.array([[58605., 546.],
                      [38853., 3423.]])

# Add premiums - percentage of expected value
# HE, deepwater, long contract, appraisal contract
rev_prems = np.array([0.15, 0.35, 0.3, 0.2])

# --- Distance between markets --- #
market_distance = np.array([['B', 'GOM', 5200.],
                            ['B', 'NS', 5700.],
                            ['B', 'WA', 2900.],
                            ['GOM', 'NS', 4600.],
                            ['GOM', 'WA', 5700.],
                            ['NS', 'WA', 5100.]
                            ], dtype=object)
# --- END INPUT --- #

if __name__ == '__main__':

    for fleet_num in range(my_fleets):
        if my_fleets == 1:
            rig_file = 'Input\RigData_Balanced.csv'
            fleet_size = len(listofRigs(rig_file))
        else:
            print 'Fleet: %s' % (str(fleet_num + 1))
            rig_file = 'Input\Fleets\\fleet%s_%sJU.csv' % (
                str(fleet_num+1), str(my_fleets-fleet_num-1))
            fleet_size = len(listofRigs(rig_file))

        # --- Results arrays --- #
        # (1) Iteration array - information about each iteration in scenario (contracts)
        it_results = np.zeros([nIterations, 6])

        # (2) Rig array - rig results for each iteration
        # 3dim array (scenario number, rig properties, rig number)
        # rig properties: profit, num.contracts served, duration/rig years, ...
        rig_results = np.zeros([nIterations, 11, fleet_size])

        for ss in range(nIterations):

            # Get Rig Objects - must be reset for each scenario
            all_rigs = listofRigs(rig_file)

            # Generate contract scenario
            # master list, used to generate arcs by indexing
            contracts, it_info = generateContracts(T, p_oil_price, fleet_size,
                                                   p_num_contracts, p_market,
                                                   p_depths, p_scope, p_duration,
                                                   p_revenue, rev_prems)
            unserved_contracts = contracts
            served_contracts = np.empty(0)

            # For each contract scenario, reset rig sets.
            existing_rigs = np.empty(0)  # existing fleet
            design_rigs = np.empty(0)  # options of rigs to acquire

            # compute costs and remove from simulation
            opex_total = [(T*12.)*i.opex for i in all_rigs]
            capex_total = [i.cost_purchase for i in all_rigs]

            for i in range(len(all_rigs)):
                all_rigs[i].opex = 0
                all_rigs[i].cost_purchase = all_rigs[i].cost_purchase + \
                    opex_total[i]

            # mode 1 -> fleet allocation, mode 0 -> rig selection
            if my_mode == 1:
                for i in all_rigs:
                    existing_rigs = np.append(
                        existing_rigs, i)  # all rigs existing
            else:  # mode 0
                for i in all_rigs:
                    design_rigs = np.append(design_rigs, i)

            active_rigs = np.empty(0)
            scrap_rigs = np.empty(0)

            # (1) ----------- Solve existing fleet allocation problem:
            while len(existing_rigs) > 0:
                # For each unsolved rig -  solve network problem
                for v in existing_rigs:
                    # Compute list of feasible nodes and get list of feasible contracts
                    nlist, cts_nlist = cs.compatibleRigs(v, unserved_contracts)
                    # Check if no feasible contracts - only start & end node
                    if len(cts_nlist) == 0:
                        # Reset previous results
                        v.profit = 0
                        v.servicingcontracts = []
                        # Remove rigs from set
                        scrap_rigs = np.append(scrap_rigs, v)
                        existing_rigs = existing_rigs[existing_rigs != v]
                    else:
                        # Solve network problem
                        ns = [lp.Node(j) for j in nlist]
                        arcs = cs.feasibleArcs(
                            ns, cts_nlist, v, market_distance)
                        nw = lp.Network(ns, arcs)
                        nw.solve(v)
                        v.profit, v.servicingcontracts = nw.result()

                # Select best rig - NB: what is selection criterion?
                # 1) select jackups, 2) select highest profit
                if len(existing_rigs) > 0:
                    num_JU = list()
                    for i in existing_rigs:
                        # Jackup Priority Criterion
                        if i.type == 'Jackup':
                            num_JU.append(i)  # record all JUs
                    if len(num_JU) > 0:  # if any Jackups
                        n = np.argmax([v.profit for v in num_JU])
                        best_rig = num_JU[n]
                    else:
                        # Return index of best rig
                        n = np.argmax([v.profit for v in existing_rigs])
                        best_rig = existing_rigs[n]

                    # Move rig to solved set
                    active_rigs = np.append(active_rigs, best_rig)
                    existing_rigs = existing_rigs[existing_rigs != best_rig]

                    # Remove contracts serviced from set
                    for i in best_rig.servicingcontracts:
                        for j in unserved_contracts:
                            if j.id == i:
                                served_contracts = np.append(
                                    served_contracts, j)
                                unserved_contracts = unserved_contracts[unserved_contracts != j]

            # (2) ----------- Solve design problem - selection:
            if my_mode == 0:
                for v in design_rigs:
                    nlist, cts_nlist = cs.compatibleRigs(v, unserved_contracts)
                    if len(cts_nlist) > 0:
                        ns = [lp.Node(j) for j in nlist]
                        arcs = cs.feasibleArcs(
                            ns, cts_nlist, v, market_distance)
                        nw = lp.Network(ns, arcs)
                        nw.solve(v)
                        v.profit, v.servicingcontracts = nw.result()

            # (3) ----------- Collect results

            # Get work specifics for rig contracts served
            # Get all contracts served allocated to respective rigs
            for i in all_rigs:
                # get all contract objects allocated to rig
                i.getContractsServed(contracts)
                i.getSumDurations()  # get total months spent on contracts
                # get number of contracts served in each market, plus transition betw. markets
                i.getSumMarkets()
                i.getDepths()  # get num.contracts by depth
                i.getScope()  # get num.contracts by scope

            # last contract may have duration after simulation end
            extra_time = np.zeros([len(all_rigs)])
            # If end time is after 2030, calculate number of months overdue
            for i in range(len(all_rigs)):
                if len(all_rigs[i].servicingcontracts) > 0:  # if any contracts
                    # last contract
                    x = contracts[all_rigs[i].servicingcontracts[0]-1].t_end
                    if x.year >= 2030:  # if end after 2030
                        extra_time[i] = x.month + (x.year-2030)*12

            # Write profit and number of contracts served for all rigs, scenario ss
            rig_results[ss] = [
                # 1. profit
                [i.profit for i in all_rigs],
                # 2. number of contracts
                [len(i.servicingcontracts) for i in all_rigs],
                # 3. number of months in work
                [i.rig_months for i in all_rigs],
                # 4. number of transitions betw. markets
                [i.market_transitions for i in all_rigs],
                # 5. number of harsh env. contracts
                [i.markets[2] for i in all_rigs],
                # 6. number of contracts that are exploration or appraisal
                [i.contract_scope[1] + i.contract_scope[2]
                 for i in all_rigs],
                # 7. number of midwater contracts
                [i.depths[1] for i in all_rigs],
                # 8. number of deepwater contracts
                [i.depths[2] for i in all_rigs],
                # 9. extra time to simulation - compute correct utilisation
                extra_time,
                # 10. operating cost
                opex_total,
                # 11. capital cost
                capex_total,
            ]

            # Write number of contracts in scenario ss
            # See ContractGeneratyor.py for contents of it_info
            it_results[ss, 0] = len(contracts)
            it_results[ss, 1] = np.sum([i.profit for i in all_rigs])
            # utilisation
            it_results[ss, 2] = np.sum(
                [i.rig_months for i in all_rigs]) / (np.sum(extra_time) + T*12.*fleet_size)
            it_results[ss, 3] = sum(capex_total) + \
                sum(opex_total)  # costs (constant)
            it_results[ss, 4:] = [i for i in it_info]

            # # (4) ----------- Display results
            print ('Iteration:     %s\n' % (str(ss+1)))
            # for i in all_rigs:
            #     print ('Unit %s serves contracts %s for a profit of %.2f. \n'
            #            % (str(i.id), str(i.servicingcontracts), i.profit))

        # ------------------------------------------------------------------------
        # (5) ----------- Post-Processing
        dict_rigs = {'Name': [i.name for i in all_rigs],
                     'HE/BE': [i.he_be for i in all_rigs],
                     'Water depth(m)': [i.water_depth for i in all_rigs],
                     }
        rig_df, scen_df = getPostProcessing(
            T, rig_results, it_results, dict_rigs)

        # (6) ----------- Write results to file
        # Select filename to save as
        if my_fleets == 1:
            if my_mode == 0:  # rig selection (1)
                filename_rigresults = 'Output\Data\IndividualRigs\\rig_result_Poil_%s.csv' % (
                    str(int(p_oil_price[0])))
                filename_scenresults = 'Output\Data\IndividualRigs\\iteration_result_Poil_%s.csv' % (
                    str(int(p_oil_price[0])))
            else:  # fleet allocation (2)
                filename_rigresults = 'Output\Data\Allocation\\rig_result_Poil_%s.csv' % (
                    str(int(p_oil_price[0])))
                filename_scenresults = 'Output\Data\Allocation\\iteration_result_Poil_%s.csv' % (
                    str(int(p_oil_price[0])))
        else:  # iterate for a set of fleets (3)
            filename_rigresults = 'Output\Data\Fleets\\rigres_fleet%s_Poil%s_mode1.csv' % (
                str(fleet_num+1), str(int(p_oil_price[0])))
            filename_scenresults = 'Output\Data\Fleets\\itres_fleet%s_Poil%s_mode1.csv' % (
                str(fleet_num+1), str(int(p_oil_price[0])))

        rig_df.to_csv(filename_rigresults, index=False)
        scen_df.to_csv(filename_scenresults, index=False)


# # Convergence study
# x = np.linspace(1, nIterations, nIterations)
# y_means = np.linspace(1, nIterations, nIterations)
# # y_means[0] = rig_results[0,0,3]
# for i in range(0, len(x)):
#     new_val = np.sum(rig_results[0:(i+1), 0, 3])
#     y_means[i] = new_val/(i+1)

# # Create horisontal lines
# upper_mean = 1.05*np.mean(rig_results[:, 0, 3])
# lower_mean = 0.95*np.mean(rig_results[:, 0, 3])

# fig, ax = plt.subplots(1, 1, figsize=(12, 7))
# ax.plot(x, rig_results[:, 0, 3], label='Profit')
# ax.plot(x, y_means, label='Moving average profit')
# ax.axhline(lower_mean, linestyle='dashed', color='k', alpha=0.5)
# ax.axhline(upper_mean, linestyle='dashed', color='k', alpha=0.5)
# ax.set(xlim=[1, 100])
# ax.legend(loc=0)
# ax.set_xlabel('Iteration', fontsize=14)
# ax.set_ylabel('Revenue ($Million)', fontsize=14)
# ax.grid(color='gray', linestyle=':')
# plt.savefig('Output\Figures\convergence_study.png')
